use chrono::NaiveDate;
use diesel::{
    deserialize::FromSql,
    prelude::*,
    serialize::ToSql,
    sql_types::Text,
    sqlite::{Sqlite, SqliteValue},
    AsExpression, FromSqlRow,
};
use serde::{Deserialize, Serialize};

use std::{fmt, str::FromStr};

pub const EVENT_TYPES: [&str; 1] = ["Watering"];

use crate::db::{schema::*, DbPool, DbResult};

#[derive(FromSqlRow, Debug, AsExpression, Serialize, Deserialize)]
#[diesel(sql_type = Text)]
pub enum EventType {
    Watering,
}

impl fmt::Display for EventType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let s = match self {
            Self::Watering => "Watering",
        };

        write!(f, "{}", s)
    }
}

#[derive(Debug)]
pub struct ParseEventTypeErr(String);

impl fmt::Display for ParseEventTypeErr {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "invalid event type '{}'", self.0)
    }
}

impl std::error::Error for ParseEventTypeErr {}

impl FromStr for EventType {
    type Err = ParseEventTypeErr;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "Watering" => Ok(Self::Watering),
            _ => Err(ParseEventTypeErr(s.to_string())),
        }
    }
}

impl FromSql<Text, Sqlite> for EventType {
    fn from_sql(bytes: SqliteValue) -> diesel::deserialize::Result<Self> {
        let s = <String as FromSql<Text, Sqlite>>::from_sql(bytes)?;

        Ok(s.as_str().parse()?)
    }
}

impl ToSql<Text, Sqlite> for EventType {
    fn to_sql<'b>(
        &'b self,
        out: &mut diesel::serialize::Output<'b, '_, Sqlite>,
    ) -> diesel::serialize::Result {
        out.set_value(self.to_string());

        Ok(diesel::serialize::IsNull::No)
    }
}

#[derive(Serialize, Queryable, Selectable)]
#[diesel(table_name = events)]
#[diesel(belongs_to(super::plant::Plant))]
#[diesel(check_for_backend(diesel::sqlite::Sqlite))]
pub struct Event {
    id: i32,
    plant_id: i32,
    event_type: EventType,
    date: NaiveDate,
    description: String,
}

#[derive(Deserialize, Insertable)]
#[diesel(table_name = events)]
#[diesel(check_for_backend(diesel::sqlite::Sqlite))]
pub struct NewEvent {
    plant_id: i32,
    event_type: EventType,
    date: NaiveDate,
    description: String,
}

impl Event {
    pub fn for_plant(pool: &DbPool, plant_id: i32) -> DbResult<Vec<Self>> {
        Ok(events::dsl::events
            .select(Self::as_select())
            .filter(events::dsl::plant_id.eq(plant_id))
            .load(&mut pool.get()?)?)
    }
}

impl NewEvent {
    pub fn insert(self, pool: &DbPool) -> DbResult<Event> {
        Ok(diesel::insert_into(events::table)
            .values(self)
            .returning(Event::as_returning())
            .get_result(&mut pool.get()?)?)
    }
}
