use axum::{
    extract::{Request, State},
    http::{HeaderMap, StatusCode},
    middleware::Next,
    response::{Html, IntoResponse, Response},
    routing::post,
    Form, Router,
};
use axum_extra::extract::{
    cookie::{Cookie, SameSite},
    CookieJar,
};
use serde::Deserialize;

use crate::db::{DbError, DbPool, Session, User};

use super::error::AppError;

pub fn logged_in_user(pool: &DbPool, headers: &HeaderMap) -> Result<Option<User>, DbError> {
    let jar = CookieJar::from_headers(headers);

    if let Some(session_id) = jar
        .get("session_id")
        .and_then(|c| c.value().parse::<i64>().ok())
    {
        Session::user_from_id(pool, session_id)
    } else {
        Ok(None)
    }
}

pub async fn auth_middleware(
    State(ctx): State<crate::Context>,
    mut req: Request,
    next: Next,
) -> Response {
    match logged_in_user(&ctx.pool, req.headers()) {
        Ok(Some(user)) => {
            req.extensions_mut().insert(user);

            next.run(req).await
        }
        Ok(None) => StatusCode::UNAUTHORIZED.into_response(),
        Err(err) => AppError::Db(err).into_response(),
    }
}

pub fn app() -> Router<crate::Context> {
    Router::new().route("/login", post(post_login))
}

#[derive(Deserialize)]
struct Login {
    username: String,
    password: String,
}

async fn post_login(
    State(ctx): State<crate::Context>,
    jar: CookieJar,
    headers: HeaderMap,
    Form(login): Form<Login>,
) -> Result<(CookieJar, Html<String>), AppError> {
    if let Some(user) = User::by_username(&ctx.pool, &login.username)? {
        if user.verify_password(&login.password) {
            let session = Session::new_for_user(&ctx.pool, user.id)?;
            let session_id = session.id.to_string();

            Ok((
                jar.add(
                    Cookie::build(("session_id", session_id))
                        .secure(true)
                        .http_only(true)
                        .same_site(SameSite::Lax)
                        .build(),
                ),
                super::render_home(ctx, &headers).await?,
            ))
        } else {
            Err(AppError::Unauthorized)
        }
    } else {
        Err(AppError::Unauthorized)
    }
}
