use std::fmt::{self, Write};

use axum::{extract::multipart::MultipartError, http::StatusCode, response::IntoResponse};

use crate::db;

#[derive(Debug)]
pub enum AppError {
    Db(db::DbError),
    Tera(tera::Error),
    Multipart(MultipartError),
    IO(std::io::Error),
    Other(Box<dyn std::error::Error + 'static + Send + Sync>),
    BadRequest,
    Unauthorized,
    NotFound,
}

impl fmt::Display for AppError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Db(_) => write!(f, "database error"),
            Self::Tera(_) => write!(f, "error rendering template"),
            Self::Multipart(_) => write!(f, "error processing multipart request"),
            Self::IO(_) => write!(f, "io error"),
            Self::Other(_) => write!(f, "other error"),
            Self::BadRequest => write!(f, "bad request"),
            Self::Unauthorized => write!(f, "unauthorized"),
            Self::NotFound => write!(f, "not found"),
        }
    }
}

impl std::error::Error for AppError {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        match self {
            Self::Db(err) => Some(err),
            Self::Tera(err) => Some(err),
            Self::IO(err) => Some(err),
            Self::Other(err) => Some(err.as_ref()),
            Self::Multipart(err) => Some(err),
            Self::NotFound | Self::Unauthorized | Self::BadRequest => None,
        }
    }
}

pub trait ErrorExt: std::error::Error {
    /// Return the full chain of error messages
    fn stack(&self) -> String {
        let mut msg = format!("{}", self);
        let mut err = self.source();

        while let Some(src) = err {
            write!(msg, " - {}", src).unwrap();

            err = src.source();
        }

        msg
    }
}

impl<E: std::error::Error> ErrorExt for E {}

impl From<db::DbError> for AppError {
    fn from(value: db::DbError) -> Self {
        Self::Db(value)
    }
}

impl From<tera::Error> for AppError {
    fn from(value: tera::Error) -> Self {
        Self::Tera(value)
    }
}

impl From<MultipartError> for AppError {
    fn from(value: MultipartError) -> Self {
        Self::Multipart(value)
    }
}

impl From<std::io::Error> for AppError {
    fn from(value: std::io::Error) -> Self {
        Self::IO(value)
    }
}

impl From<image::ImageError> for AppError {
    fn from(value: image::ImageError) -> Self {
        match value {
            image::ImageError::IoError(err) => Self::IO(err),
            _ => Self::BadRequest,
        }
    }
}

impl IntoResponse for AppError {
    fn into_response(self) -> axum::response::Response {
        match self {
            Self::NotFound => StatusCode::NOT_FOUND.into_response(),
            Self::Unauthorized => StatusCode::UNAUTHORIZED.into_response(),
            Self::BadRequest => StatusCode::BAD_REQUEST.into_response(),
            Self::Multipart(err) => err.into_response(),
            _ => {
                tracing::error!("{}", self.stack());

                StatusCode::INTERNAL_SERVER_ERROR.into_response()
            }
        }
    }
}
