use axum::{
    extract::{DefaultBodyLimit, Multipart, Path, Request, State},
    handler::Handler,
    response::{Html, IntoResponse},
    routing::{get, post},
    Router,
};
use chrono::NaiveDate;
use futures::TryStreamExt;
use image::{codecs::jpeg::JpegEncoder, ImageReader};
use mime::Mime;
use tokio_util::io::StreamReader;
use tower::ServiceExt;
use tower_http::services::ServeFile;

use std::{io::BufWriter, path::PathBuf};

use super::error::AppError;
use crate::{
    db::{Image, NewImage},
    IMG_DIR,
};

const THUMBNAIL_EXT: &str = "thumb";

pub fn app() -> axum::Router<crate::Context> {
    Router::new()
        .route(
            "/",
            post(post_image.layer(DefaultBodyLimit::max(1024 * 1024 * 20))),
        )
        .route("/{id}/original", get(get_image_original))
        .route("/{id}/thumb", get(get_image_thumb))
}

async fn get_image(
    ctx: crate::Context,
    id: i32,
    req: Request,
    thumb: bool,
) -> super::Result<impl IntoResponse> {
    let image = tokio::task::spawn_blocking(move || Image::by_id(&ctx.pool, id))
        .await
        .unwrap()?
        .ok_or(AppError::NotFound)?;

    let mime: Mime = image
        .mime_type
        .parse()
        .map_err(|err| AppError::Other(Box::new(err)))?;

    let mut path = ctx.data_dir.join(crate::IMG_DIR).join(image.id.to_string());

    if thumb {
        path.set_extension(THUMBNAIL_EXT);
    }

    Ok(ServeFile::new_with_mime(path, &mime).oneshot(req).await)
}

async fn get_image_original(
    State(ctx): State<crate::Context>,
    Path(id): Path<i32>,
    req: Request,
) -> super::Result<impl IntoResponse> {
    get_image(ctx, id, req, false).await
}

async fn get_image_thumb(
    State(ctx): State<crate::Context>,
    Path(id): Path<i32>,
    req: Request,
) -> super::Result<impl IntoResponse> {
    get_image(ctx, id, req, true).await
}

async fn post_image(
    State(ctx): State<crate::Context>,
    mut mt: Multipart,
) -> super::Result<Html<String>> {
    let mut plant_id: Option<i32> = None;
    let mut date_taken: Option<NaiveDate> = None;
    let mut note: Option<String> = None;

    let mut img_paths: Option<(PathBuf, PathBuf)> = None;
    let mut mime_type: Option<&'static str> = None;

    while let Some(field) = mt.next_field().await? {
        match field.name() {
            Some("plant_id") => {
                plant_id = Some(
                    field
                        .text()
                        .await?
                        .parse()
                        .map_err(|_| AppError::BadRequest)?,
                );
            }
            Some("date_taken") => {
                date_taken = Some(
                    field
                        .text()
                        .await?
                        .parse()
                        .map_err(|_| AppError::BadRequest)?,
                );
            }
            Some("note") => {
                note = Some(field.text().await?);
            }
            Some("image") => {
                let path = ctx.tmp_file_path().await;

                let mut f = tokio::fs::File::create(&path).await?;
                let mut r = StreamReader::new(field.map_err(std::io::Error::other));

                tokio::io::copy(&mut r, &mut f).await?;

                let jpeg_path = ctx.tmp_file_path().await;
                img_paths = Some((path.clone(), jpeg_path.clone()));

                mime_type = Some(
                    tokio::task::spawn_blocking(move || process_image(path, jpeg_path))
                        .await
                        .unwrap()?,
                );
            }
            _ => {
                return Err(AppError::BadRequest);
            }
        }
    }

    if let (Some(plant_id), Some(date_taken), Some((img_path, jpeg_path)), Some(mime_type)) =
        (plant_id, date_taken, img_paths, mime_type)
    {
        let image = NewImage::new(plant_id, date_taken, mime_type.to_string(), note);
        let image =
            tokio::task::spawn_blocking(move || insert_image(&ctx, image, img_path, jpeg_path))
                .await
                .unwrap()?;

        Ok(Html(String::new()))
    } else {
        Err(AppError::BadRequest)
    }
}

fn process_image(path: PathBuf, jpg_path: PathBuf) -> super::Result<&'static str> {
    let reader = ImageReader::open(path)?.with_guessed_format()?;
    let mime_type = reader.format().ok_or(AppError::BadRequest)?.to_mime_type();

    // Decode image and resize to fit into 720 x 720 square
    let img = reader.decode()?.thumbnail(720, 720);

    // Write image as compressed JPEG
    let jpeg_f = BufWriter::new(std::fs::File::create(jpg_path)?);
    let encoder = JpegEncoder::new_with_quality(jpeg_f, 60);
    img.write_with_encoder(encoder)?;

    Ok(mime_type)
}

fn insert_image(
    ctx: &crate::Context,
    image: NewImage,
    img_path: PathBuf,
    jpeg_path: PathBuf,
) -> super::Result<Image> {
    let image = image.insert(&ctx.pool)?;

    let dest_path = ctx.data_dir.join(IMG_DIR).join(image.id.to_string());
    std::fs::rename(img_path, &dest_path)?;
    std::fs::rename(jpeg_path, dest_path.with_extension(THUMBNAIL_EXT))?;

    Ok(image)
}
