use axum::{
    extract::{Path, State},
    http::HeaderMap,
    response::Html,
    routing::{get, post},
    Form, Router,
};
use tera::Context;

use crate::db::{self, DbError, Event, Plant};

use super::error::AppError;

pub fn app() -> axum::Router<crate::Context> {
    Router::new()
        .route("/{id}", get(get_plant_page))
        .route("/", post(post_plant))
}

async fn get_plant_page(
    State(ctx): State<crate::Context>,
    headers: HeaderMap,
    Path(plant_id): Path<i32>,
) -> super::Result<Html<String>> {
    let res = tokio::task::spawn_blocking(move || {
        let plant = Plant::by_id(&ctx.pool, plant_id)?;

        if let Some(plant) = plant {
            let events = Event::for_plant(&ctx.pool, plant.id)?;

            Ok::<_, DbError>(Some((plant, events)))
        } else {
            Ok(None)
        }
    })
    .await
    .unwrap()?;

    match res {
        Some((plant, events)) => {
            let mut context = Context::new();
            context.insert("plant", &plant);
            context.insert("events", &events);
            context.insert("event_types", &db::EVENT_TYPES);

            Ok(Html(super::render_view(
                &ctx.tera,
                "views/plant.html",
                &context,
                &headers,
            )?))
        }
        None => Err(AppError::NotFound),
    }
}

async fn post_plant(
    State(ctx): State<crate::Context>,
    Form(plant): Form<db::NewPlant>,
) -> super::Result<Html<String>> {
    let plant = tokio::task::spawn_blocking(move || plant.insert(&ctx.pool))
        .await
        .unwrap()?;

    let mut context = Context::new();
    context.insert("plant", &plant);
    Ok(Html(ctx.tera.render("updates/plant_li.html", &context)?))
}
